/* Inspired by Yinliang Wu's blog "Solve the Missionaries and Cannibals Problem with SAS"
 * at https://blogs.sas.com/content/sgf/2022/02/15/missionaries-and-cannibals-problem/ 
 * Author: Rob Pratt
 */
option nosource;
%macro solve(M=, C=, B=);
   proc optmodel;
      /* <m,c,b> is number of missionaries, cannibals, and boats on the left bank */
      set STATES = {m in 0..&M, c in 0..&C, b in 0..1: (m = 0 or m >= c) and (&M - m = 0 or &M - m >= &C - c)};

      /* <m,c> is number of missionaries and cannibals on the boat */
      set BOAT = {m in 0..&M, c in 0..&C: (m = 0 or m >= c) and 1 <= m + c <= &B};

      /* collapse STATES to character NODES */
      set <str> NODES init {};
      str nodeThis;
      str nodeId {STATES};
      for {<m,c,b> in STATES} do;
         nodeThis = m||','||c||','||b;
         NODES = NODES union {nodeThis};
         nodeId[m,c,b] = nodeThis;
      end;

      /* construct LINKS <i,j> from node i to node j */
      set LINKS init
         /* left bank to right bank */
         setof {<m,c,1> in STATES, <mb,cb> in BOAT: <m-mb,c-cb,0> in STATES} <nodeId[m,c,1],nodeId[m-mb,c-cb,0]>
            union
         /* right bank to left bank */
         setof {<m,c,0> in STATES, <mb,cb> in BOAT: <m+mb,c+cb,1> in STATES} <nodeId[m,c,0],nodeId[m+mb,c+cb,1]>;

      /* start with everybody on the left bank */
      str source = nodeId[&M,&C,1];
      /* finish with nobody on the left bank */
      str sink   = nodeId[0,0,0];
      set SOURCES = {source};
      set SINKS   = {sink};
      set <str,str,num,num,str>     PATHSNODES;   /* <source,sink,path,order,node> */
      set <str,str,num,num,str,str> PATHSLINKS;   /* <source,sink,path,order,from,to> */
      set <str,str,num,num>         PATHSWEIGHTS; /* <source,sink,path,length> */

      /* disallow back-links that decrease distance from source */
      num spweights {SOURCES, NODES} init .;
      solve with network / shortpath=(source=SOURCES) direction=directed links=(include=LINKS)
         out=(spweights=spweights);
      LINKS = {<i,j> in LINKS: j = sink or spweights[source,i] <= spweights[source,j]};

      /* find all paths from source to sink */
      solve with network / path=(source=SOURCES sink=SINKS) direction=directed links=(include=LINKS)
         out=(pathsnodes=PATHSNODES pathslinks=PATHSLINKS);

      /* save union of path links to data set */
      create data graph&M._&C._&B from [from next]=(setof {<(source),(sink),path,order,i,j> in PATHSLINKS} <i,j>);

      /* save path lengths to data set */
      set PATHS = 1.._OROPTMODEL_NUM_['NUM_PATHS'];
      num length {PATHS} init 0;
      for {<(source),(sink),path,order,from,next> in PATHSLINKS} length[path] = length[path] + 1;
      create data path&M._&C._&B from [path] length;
   quit;

   /* summarize path lengths */
   proc sql;
      create table summary&M._&C._&B as
      select min(length) as MinTrips, max(length) as MaxTrips, count(distinct length) as DistinctTripsLength
      from path&M._&C._&B;
   quit;
%mend solve;

%macro NetDrawEx(M=3, C=3, B=2, ds=graph&M._&C._&B., Opt=%str());
   %if (%eval(&B*&C) le 4) OR (%eval(&M/&B) eq 2) OR (&M le &B) %then %let Opt=%str(hMARGIN=12 vMARGIN=10);
   %if %sysfunc(exist(&ds)) %then %do;   
      data _null_; if 0 then set &ds nobs=n; call symputx('nrows',n);stop;run; /* exists but number of observation is 0 */
      %if &NROWS = 0 %then %return;
      proc netdraw graphics data=&ds out=&ds._layout;
         title "Missionaries and Cannibals Problem (M=&M C=&C B=&B)";
         actnet / act=from succ=(next) 
            cnodefill=white coutline=white 
            carcs=black  
            ctext=black centerid compress FILLPAGE &opt
            spanningtree font="SimSun/Bold" htext=2 arrowhead=2;
      run;
      title;
   %end;
%mend NetDrawEx;

%solve(M=3, C=3, B=2)
%NetDrawEx(M=3, C=3, B=2)
